import numpy as np
import xarray as xr
import cftime
import pandas as pd
from scipy.optimize import leastsq
import os
import matplotlib.pyplot as plt
import cftime
from domino.util import xarr_times_to_ints

class Agg_Deseasonaliser(object):
    """
    Computes a seasonal cycle from an xarray dataset based on an aggregation
    criteria i.e. month, day of year, week of year etc.
    ----------
    Methods
    -------
    fit_cycle(DataArray,dim=str,agg=str): Aggregates the input DataArray arr.

    evaluate_cycle(data=DataArray): When data is None, creates a seasonal cycle
    DataArray the same shape as the input data used in fit_cycle. data can be
    a DataArray with a differing dim coordinate, but matching other coordinates.
    This can be used to fit a seasonal cycle on one dataset, but evaluate it on
    another, as long as the grid, levels, etc are the same.
    Returns The seasonal cycle as a DataArray
    """
    def __init__(self):
        self.data=None
        self.cycle_coeffs=None


    def fit_cycle(self,arr,dim="time",agg="dayofyear"):
        var=dim+"."+agg
        self.data=arr
        self.dim=dim
        self.agg=agg
        self.cycle_coeffs=arr.groupby(var).mean(dim=dim)
        return

    def evaluate_cycle(self,data=None,smooth=1):
        if data is None:
            data=self.data
        cycle=self.cycle_coeffs.sel({self.agg:getattr(data[self.dim].dt,self.agg).data})
        return self.smooth(cycle,smooth)
    
    def smooth(self,arr,w):
        if w==1:
            return arr
        else:
            return arr.rolling({self.agg:w},min_periods=1,center=True).mean()
        
class Agg_FlexiDeseasonaliser(Agg_Deseasonaliser):
    """Modifies Agg_Deseasonaliser, with the ability
    to handle custom summary funcs other than the mean.
    summary_func must take an xarray.groupby object, and
    accept a keyword 'dim'."""
    
    def fit_cycle(self,arr,summary_func,dim="time",agg="dayofyear"):
        var=dim+"."+agg
        self.data=arr
        self.dim=dim
        self.agg=agg
        self.cycle_coeffs=summary_func(arr.groupby(var),dim=dim)
        return

    
class Sinefit_Deseasonaliser(Agg_Deseasonaliser):

    """
    Same functionalityy as Agg_Deseasonaliser but fits a number of sin modes
    to the data using least squares fitting. This produces a smoother cycle than
    aggregating but can be slow for large datasets.
    ----------
    Methods
    -------
    fit_cycle(DataArray,dim=str,N=int,period=float): N sets the number of sin modes
    to fit as fractions of the period. i.e. the default of N=4 fits waves with
    periods of period, period/2, period/3 and period/4. period is a float setting
    the length of a year in days.

    evaluate_cycle(data=DataArray): As for Agg_Deseasonaliser
    """
    def _func_residual(self,p,x,t,N,period):
        return x - self._evaluate_fit(t,p,N,period)

    def _evaluate_fit(self,t,p,N,period):
        ans=p[1]
        for i in range(0,N):
            ans+=p[2*i+2] * np.sin(2 * np.pi * (i+1)/period * t + p[2*i+3])
        return ans

    def _lstsq_sine_fit(self,arr,t,N,period):

        #Guess initial parameters
        std=arr.std()
        p=np.zeros(2*(N+1))
        for i in range(0,N):
            p[2+2*i]=std/(i+1.0)
        plsq=leastsq(self._func_residual,p,args=(arr,t,N,period))
        return plsq[0]

    def fit_cycle(self,arr,dim="time",N=4,period=365.25):

            dims=arr.dims
            self.dim=dim
            self.data=arr
            self.N=N
            self.period=period
            t=xarr_times_to_ints(arr[dim])
            self.coeffs= xr.apply_ufunc(
                self._lstsq_sine_fit,
                arr,
                input_core_dims=[[dim]],
                output_core_dims=[["coeffs"]],
                vectorize=True,
                kwargs={"t":t,"N":N,"period":period})
            return

    def evaluate_cycle(self,data=None):
        if data is None:
            data=self.data
        dims=data.dims
        t=xarr_times_to_ints(data[self.dim])
        print(self.coeffs.shape)
        cycle=np.array([[self._evaluate_fit(t,c2.data,self.N,self.period)\
                for c2 in c1] for c1 in np.atleast_3d(self.coeffs)])

        return data.transpose(...,"time").copy(data=cycle).transpose(*dims)



from domino.categorical_analysis import get_xr_occurrence

def aggregate_ds(ds,dim,agg_func,**agg_kwargs):
    return ds.map(agg_func,dim=dim,**agg_kwargs)
def mean_ds(ds,dim):
    return aggregate_ds(ds,dim,lambda da,dim: da.mean(dim=dim))
def std_ds(ds,dim):
    return aggregate_ds(ds,dim,lambda da,dim: da.std(dim=dim))
def cat_occ_ds(ds,dim,cat_ds):
    return aggregate_ds(ds,dim,get_xr_occurrence,s=cat_ds,coord_name='variable_cat_val')



import numpy as np
import xarray as xr
import operator as op
from functools import partial

#Set the object up, pass in a mask, and get a filtered mask back
class MaskFilter(object):
    
    def __init__(self):
        
        self._filter_funcs=[]
        self._required_kwargs=[]

        self.filter_description=[]
        self.long_filter_description=[]
        
        
        self._filter_modes=['union','intersection','replace']
        
        self._comparators={
            'lt':op.lt,
            'gt':op.gt,
            'le':op.le,
            'ge':op.ge,
            'eq':op.eq,
            'neq':op.ne}
        return
    
    
    #valid inputs:
    #        an integer (1 value for everything)
    #        a dict of var names and integers (1 value per var)
    #        a dataset with values on each coord
    
    def add_value_filter(self,threshold, comparison_func, filter_mode='union',as_abs=False):
        
        assert comparison_func in self._comparators
        assert filter_mode in self._filter_modes
        
        comparator=self._comparators[comparison_func]
        if as_abs:
            comparator=lambda val_ds,thresh_ds: self._comparators[comparison_func](np.abs(val_ds),thresh_ds)
            
        value_filter=partial(self._apply_value_filter,threshold,comparator,filter_mode)
        
        self._filter_funcs.append(value_filter)
        self._required_kwargs.append('val_ds')
        self.filter_description.append('value filter')
        self.long_filter_description.append(\
            f'value filter, comparator: {comparison_func}, filter_mode: {filter_mode}, as_abs: {as_abs}')
        
        return
    def _apply_value_filter(self,threshold,comparator,filter_mode,mask_ds=None,val_ds=None):
        
        vfilt=comparator(val_ds,threshold)
        
        if filter_mode=='union':
            vfilt=vfilt or mask_ds
        elif filter_mode=='intersection':
            vfilt=vfilt*mask_ds
        else:
            assert filter_mode=='replace'
            
        return vfilt
    
    def add_convolution_filter():
        return
    def _apply_convolution_filter():
        return

    def add_continuity_filter():
        return
    
    
    def _apply_continuity_filter():
        return

    def filter_mask(self,mask_ds,**kwargs):
        mask_ds=mask_ds.copy(deep=True)
        for f,fname in zip(self._filter_funcs,self.filter_description):
            try:
                mask_ds=f(mask_ds,**kwargs)
            except Exception as e:
                raise(ValueError(f'Failure applying filter {fname}:\n{e}'))
        return mask_ds
        
    
        
class IndexComputer(object):
    """x
    
        **Arguments:**
        
        *a*
            xxx
            
        **Optional arguments**
        
        *a*
        
            xxx
"""
    def __init__(self):
        
        return
    
    def __repr__(self):
        return 1
    
    def __str__(self):
        return self.__repr__()
